@@ -10,6 +10,7 @@
 #include "elaborate.h"
 #include "parse.h"
 #include "tapsets.h"
+#include "session.h"
 
 extern "C" {
 #include <sys/utsname.h>
@@ -69,6 +70,65 @@ derived_probe::derived_probe (probe *p, probe_point *l):
 }
 
 
+// ------------------------------------------------------------------------
+// Members of derived_probe_group
+
+void
+derived_probe_group::register_probe(be_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a be_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(dwarf_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a dwarf_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(hrtimer_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a hrtimer_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(mark_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a mark_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(never_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a never_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(profile_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a profile_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(timer_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a timer_derived_probe");
+}
+
+
+void
+derived_probe_group::register_probe(perfmon_derived_probe* p)
+{
+  throw semantic_error ("unexpected registration of a perfmon_derived_probe");
+}
+
+
 // ------------------------------------------------------------------------
 // Members of derived_probe_builder
 
@@ -330,9 +390,12 @@ struct alias_derived_probe: public derived_probe
 {
   alias_derived_probe (probe* base): derived_probe (base) {}
 
+  void register_probe (systemtap_session& s) { }
+
   // alias probes should be ultimately expanded to other derived_probe
   // types, and not themselves emitted.
-  void emit_registrations (translator_output* o) { throw semantic_error ("inappropriate", this->tok); }
+  void emit_registrations_start (translator_output* o, unsigned index) { throw semantic_error ("inappropriate", this->tok); }
+  void emit_registrations_end (translator_output* o, unsigned index) { throw semantic_error ("inappropriate", this->tok); }
   void emit_deregistrations (translator_output* o) { throw semantic_error ("inappropriate", this->tok); }
   void emit_probe_entries (translator_output* o) { throw semantic_error ("inappropriate", this->tok); }
 };
@@ -812,7 +875,8 @@ semantic_pass_stats (systemtap_session & sess)
 
 
 static int semantic_pass_symbols (systemtap_session&);
-static int semantic_pass_optimize (systemtap_session&);
+static int semantic_pass_optimize1 (systemtap_session&);
+static int semantic_pass_optimize2 (systemtap_session&);
 static int semantic_pass_types (systemtap_session&);
 static int semantic_pass_vars (systemtap_session&);
 static int semantic_pass_stats (systemtap_session&);
@@ -881,7 +945,7 @@ semantic_pass_symbols (systemtap_session& s)
           for (unsigned j=0; j<dps.size(); j++)
             {
               derived_probe* dp = dps[j];
-	      s.probes.push_back (dp);
+	      dp->register_probe (s);
 
               try 
                 {
@@ -913,8 +977,9 @@ semantic_pass (systemtap_session& s)
       register_standard_tapsets(s);
       
       rc = semantic_pass_symbols (s);
-      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize (s);
+      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize1 (s);
       if (rc == 0) rc = semantic_pass_types (s);
+      if (rc == 0 && ! s.unoptimized) rc = semantic_pass_optimize2 (s);
       if (rc == 0) rc = semantic_pass_vars (s);
       if (rc == 0) rc = semantic_pass_stats (s);
     }
@@ -932,7 +997,10 @@ semantic_pass (systemtap_session& s)
 
 systemtap_session::systemtap_session ():
   pattern_root(new match_node),
-  user_file (0), op (0), up (0), num_errors (0)
+  user_file (0), op (0), up (0),
+  kprobes_text_initialized (false),
+  kprobes_text_start (0), kprobes_text_end (0),
+  num_errors (0)
 {
 }
 
@@ -1550,9 +1618,104 @@ void semantic_pass_opt4 (systemtap_session& s, bool& relaxed_p)
     }
 }
 
+struct duplicate_function_remover: public functioncall_traversing_visitor
+{
+  systemtap_session& s;
+  map<functiondecl*, functiondecl*>& duplicate_function_map;
+
+  duplicate_function_remover(systemtap_session& sess,
+			     map<functiondecl*, functiondecl*>&dfm):
+    s(sess), duplicate_function_map(dfm) {};
+
+  void visit_functioncall (functioncall* e);
+};
+
+void
+duplicate_function_remover::visit_functioncall (functioncall *e)
+{
+  functioncall_traversing_visitor::visit_functioncall (e);
+
+  // If the current function call reference points to a function that
+  // is a duplicate, replace it.
+  if (duplicate_function_map.count(e->referent) != 0)
+    {
+      if (s.verbose>2)
+	  clog << "Changing " << e->referent->name
+	       << " reference to "
+	       << duplicate_function_map[e->referent]->name
+	       << " reference\n";
+      e->tok = duplicate_function_map[e->referent]->tok;
+      e->function = duplicate_function_map[e->referent]->name;
+      e->referent = duplicate_function_map[e->referent];
+    }
+}
+
+static string
+get_functionsig (functiondecl* f)
+{
+  ostringstream s;
+
+  // Get the "name:args body" of the function in s.  We have to
+  // include the args since the function 'x1(a, b)' is different than
+  // the function 'x2(b, a)' even if the bodies of the two functions
+  // are exactly the same.
+  f->printsig(s);
+  f->body->print(s);
+
+  // printsig puts f->name + ':' on the front.  Remove this
+  // (otherwise, functions would never compare equal).
+  string str = s.str().erase(0, f->name.size() + 1);
+
+  // Return the function signature.
+  return str;
+}
+
+void semantic_pass_opt5 (systemtap_session& s, bool& relaxed_p)
+{
+  // Walk through all the functions, looking for duplicates.
+  map<string, functiondecl*> functionsig_map;
+  map<functiondecl*, functiondecl*> duplicate_function_map;
+  for (unsigned i=0; i < s.functions.size(); /* see below */)
+    {
+      string functionsig = get_functionsig(s.functions[i]);
+
+      if (functionsig_map.count(functionsig) == 0)
+	{
+	  // This function is unique.  Remember it.
+	  functionsig_map[functionsig] = s.functions[i];
+	  i++;
+	}
+      else
+        {
+	  // This function is a duplicate.
+	  duplicate_function_map[s.functions[i]]
+	    = functionsig_map[functionsig];
+
+	  // Remove the duplicate function (since we don't need it
+	  // anymore).
+	  s.functions.erase (s.functions.begin() + i);
+
+	  relaxed_p = false;
+          // NB: don't increment i
+	}
+    }
+
+  // If we have duplicate functions, traverse down the tree, replacing
+  // the appropriate function calls.
+  // duplicate_function_remover::visit_functioncall() handles the
+  // details of replacing the function calls.
+  if (duplicate_function_map.size() != 0)
+    {
+      duplicate_function_remover dfr (s, duplicate_function_map);
+
+      for (unsigned i=0; i < s.probes.size(); i++)
+	s.probes[i]->body->visit(&dfr);
+    }
+}
+
 
 static int
-semantic_pass_optimize (systemtap_session& s)
+semantic_pass_optimize1 (systemtap_session& s)
 {
   // In this pass, we attempt to rewrite probe/function bodies to
   // eliminate some blatantly unnecessary code.  This is run before
@@ -1583,6 +1746,33 @@ semantic_pass_optimize (systemtap_session& s)
 }
 
 
+static int
+semantic_pass_optimize2 (systemtap_session& s)
+{
+  // This is run after type inference.  We run an outer "relaxation"
+  // loop that repeats the optimizations until none of them find
+  // anything to remove.
+
+  int rc = 0;
+
+  bool relaxed_p = false;
+  while (! relaxed_p)
+    {
+      relaxed_p = true; // until proven otherwise
+
+      semantic_pass_opt5 (s, relaxed_p);
+    }
+
+  if (s.probes.size() == 0)
+    {
+      cerr << "semantic error: no probes found." << endl;
+      rc = 1;
+    }
+
+  return rc;
+}
+
+
 
 // ------------------------------------------------------------------------
 // type resolution
@@ -1994,7 +2184,10 @@ typeresolution_info::visit_target_symbol (target_symbol* e)
   // later unused-expression-elimination pass didn't get rid of it
   // either.  So we have a target symbol that is believed to be of
   // genuine use, yet unresolved by the provider.
-  throw semantic_error("unresolved target-symbol expression", e->tok);
+  if (e->saved_conversion_error)
+    throw (* (e->saved_conversion_error));
+  else
+    throw semantic_error("unresolved target-symbol expression", e->tok);
 }
 
 
