@@ -162,11 +162,7 @@ bool eval_pp_conditional (systemtap_session& s,
                                     l->content == "kernel_vr"))
     {
       string target_kernel_vr = s.kernel_release;
-      string target_kernel_v = target_kernel_vr;
-      // cut off any release code suffix
-      string::size_type dr = target_kernel_vr.find ('-');
-      if (dr > 0 && dr != string::npos)
-        target_kernel_v = target_kernel_vr.substr (0, dr);
+      string target_kernel_v = s.kernel_base_release;
       
       if (! (r->type == tok_string))
         throw parse_error ("expected string literal", r);
@@ -374,6 +370,17 @@ parser::expect_unknown (token_type tt, string & target)
 }
 
 
+const token* 
+parser::expect_unknown2 (token_type tt1, token_type tt2, string & target)
+{
+  const token *t = next();
+  if (!(t && (t->type == tt1 || t->type == tt2)))
+    throw parse_error ("expected " + tt2str(tt1) + " or " + tt2str(tt2));
+  target = t->content;
+  return t;
+}
+
+
 const token* 
 parser::expect_op (std::string const & expected)
 {
@@ -405,6 +412,13 @@ parser::expect_ident (std::string & target)
 }
 
 
+const token* 
+parser::expect_ident_or_keyword (std::string & target)
+{
+  return expect_unknown2 (tok_identifier, tok_keyword, target);
+}
+
+
 bool 
 parser::peek_op (std::string const & op)
 {
@@ -767,7 +781,7 @@ parser::parse ()
 	  else if (t->type == tok_keyword && t->content == "global")
 	    {
 	      context = con_global;
-	      parse_global (f->globals);
+	      parse_global (f->globals, f->probes);
 	    }
 	  else if (t->type == tok_keyword && t->content == "function")
 	    {
@@ -1006,7 +1020,7 @@ parser::parse_statement ()
 
 
 void
-parser::parse_global (vector <vardecl*>& globals)
+parser::parse_global (vector <vardecl*>& globals, vector<probe*>& probes)
 {
   const token* t0 = next ();
   if (! (t0->type == tok_keyword && t0->content == "global"))
@@ -1028,7 +1042,56 @@ parser::parse_global (vector <vardecl*>& globals)
       globals.push_back (d);
 
       t = peek ();
-      if (t && t->type == tok_operator && t->content == ",")
+      if (t && t->type == tok_operator && t->content == "=") // initialization
+        {
+          next ();
+
+          // Create synthetic "begin" probe to assign value to global.
+          //
+          // An alternative would be to store the initializer value
+          // within a new field of vardecl, and use e.g. the
+          // type-resolution stage to do the rewriting.  It could go
+          // farther: the initializer could live through till
+          // translation and be emitted as a plain C-level variable
+          // initializer.
+
+          literal* value = parse_literal ();
+
+          probe_point* pp = new probe_point;
+          probe_point::component* ppc = new probe_point::component;
+          ppc->functor = string("begin");
+          pp->tok = t;
+          pp->components.push_back (ppc);
+
+          probe* p = new probe;
+          p->tok = t;
+          p->locations.push_back (pp);
+
+          symbol* sym = new symbol;
+          sym->tok = t;
+          sym->name = d->name;
+
+          assignment* a = new assignment;
+          a->tok = t;
+          a->op = "=";
+          a->left = sym;
+          a->right = value;
+
+          expr_statement* es = new expr_statement;
+          es->tok = t;
+          es->value = a;
+
+          block* blk = new block;
+          blk->tok = t;
+          blk->statements.push_back (es);
+
+          p->body = blk;
+          probes.push_back (p);
+
+          t = peek ();
+        }
+
+      if (t && t->type == tok_operator && t->content == ",") // next global
 	{
 	  next ();
 	  continue;
@@ -2172,7 +2235,7 @@ parser::parse_symbol ()
 	      if (peek_op ("->"))
 		{ 
 		  next(); 
-		  expect_ident (c);
+		  expect_ident_or_keyword (c);
 		  tsym->components.push_back
 		    (make_pair (target_symbol::comp_struct_member, c));
 		}
